#include <asm/types.h>
#include <stdlib.h>
#include <syslog.h>
#include <time.h>
#include "usefull.h"
#include "packet80211b.h"
#include "ap_packet.h"
#include "ap_stat.h"


extern ap_stat_elem *ap_stat_list;
extern ap_packet_elem *ap_packet_list;
extern int syslog_alert;

void
flush_ap_stat ()
{
  ap_stat_elem *current;
  current = ap_stat_list;

  while (ap_stat_list != NULL)
    {
      current = ap_stat_list;
      ap_stat_list = ap_stat_list->next;
      free (current);
    }
}

void
dump_ap_stat ()
{
  ap_stat_elem *current;
  char *buf1;
  char *buf2;
  char *time1;
  char *time2;

  current = ap_stat_list;
  buf1 = (char *) malloc (30 * sizeof (char));
  buf2 = (char *) malloc (30 * sizeof (char));


  while (current != NULL)
    {
      /* REENTRANT safe */
      time1 = (char *) ctime_r (&current->data.t_first_packet.tv_sec, buf1);
      time2 = (char *) ctime_r (&current->data.t_last_packet.tv_sec, buf2);
      printf
	("\n-----------------\nAP mac : %s\ndst mac : %s\ntype :0x%02x\npackets : %d\nfirst : %slast : %s",
	 human_format_hw_addr (current->data.ap_mac),
	 human_format_hw_addr (current->data.dst_mac), current->data.type,
	 current->data.num_offending_packet, time1, time2);
      current = current->next;
    }
  free (buf1);
  free (buf2);
}

void
msg_alert (float debit, ap_stat_elem * alert)
{

  char *msgwho;
  char msgcard[30];
  char msgdetails[100];
  struct tm *tim1;
  struct tm *tim2;

  tim1 = (struct tm *) malloc (sizeof (struct tm));
  tim2 = (struct tm *) malloc (sizeof (struct tm));

  msgwho = (char *) malloc (30 * sizeof (char));
  sprintf (msgwho, "AP %s", human_format_hw_addr (alert->data.ap_mac));
  if (isbroadcast (alert->data.dst_mac))
    strcpy (msgcard, "broadcast");
  else
    sprintf (msgcard, "card %s", human_format_hw_addr (alert->data.dst_mac));

  /* REENTRANT safe */
  tim1 = (struct tm *) localtime_r (&alert->data.t_first_packet.tv_sec, tim1);
  tim2 = (struct tm *) localtime_r (&alert->data.t_last_packet.tv_sec, tim2);

  switch (alert->data.type)
    {
    case 0x10:			/* auth response */
      if (debit > MAX_NUM_AUTH_RESP)
	{
	  sprintf (msgdetails,
		   "(%d authentication responses send to %s between %02d:%02d:%02d and %02d:%02d:%02d)",
		   alert->data.num_offending_packet, msgcard, tim1->tm_hour,
		   tim1->tm_min, tim1->tm_sec, tim2->tm_hour, tim2->tm_min,
		   tim2->tm_sec);
	  if (syslog_alert == 1)
	    syslog (LOG_ALERT, "%s %s", msgwho, msgdetails);
	  else
	    printf ("\n%s %s\n", msgwho, msgdetails);
	}
      break;

    case 0x30:			/* assoc response */
      if (debit > MAX_NUM_ASSOC_RESP)
	{
	  sprintf (msgdetails,
		   "(%d association responses send to %s between %02d:%02d:%02d and %02d:%02d:%02d)",
		   alert->data.num_offending_packet, msgcard, tim1->tm_hour,
		   tim1->tm_min, tim1->tm_sec, tim2->tm_hour, tim2->tm_min,
		   tim2->tm_sec);
	  if (syslog_alert == 1)
	    syslog (LOG_ALERT, "%s %s", msgwho, msgdetails);
	  else
	    printf ("\n%s %s\n", msgwho, msgdetails);
	}
      break;

    case 0x50:			/* probe response */
      if (debit > MAX_NUM_PROBE_RESP)
	{
	  sprintf (msgdetails,
		   "(%d probe responses send to %s between %02d:%02d:%02d and %02d:%02d:%02d)",
		   alert->data.num_offending_packet, msgcard, tim1->tm_hour,
		   tim1->tm_min, tim1->tm_sec, tim2->tm_hour, tim2->tm_min,
		   tim2->tm_sec);
	  if (syslog_alert == 1)
	    syslog (LOG_ALERT, "%s %s", msgwho, msgdetails);
	  else
	    printf ("\n%s %s\n", msgwho, msgdetails);
	}
      break;

    case 0xa0:			/* desassociacion */
      if (debit > MAX_NUM_DISASSOC)
	{
	  sprintf (msgdetails,
		   "(%d desassociation packets send to %s between %02d:%02d:%02d and %02d:%02d:%02d)",
		   alert->data.num_offending_packet, msgcard, tim1->tm_hour,
		   tim1->tm_min, tim1->tm_sec, tim2->tm_hour, tim2->tm_min,
		   tim2->tm_sec);
	  if (syslog_alert == 1)
	    syslog (LOG_ALERT, "%s %s", msgwho, msgdetails);
	  else
	    printf ("\n%s %s\n", msgwho, msgdetails);
	}
      break;

    case 0xc0:			/* deauthentication */
      if (debit > MAX_NUM_DEAUTH)
	{
	  sprintf (msgdetails,
		   "(%d deauthentication packets send to %s between %02d:%02d:%02d and %02d:%02d:%02d)",
		   alert->data.num_offending_packet, msgcard, tim1->tm_hour,
		   tim1->tm_min, tim1->tm_sec, tim2->tm_hour, tim2->tm_min,
		   tim2->tm_sec);
	  if (syslog_alert == 1)
	    syslog (LOG_ALERT, "%s %s", msgwho, msgdetails);
	  else
	    printf ("\n%s %s\n", msgwho, msgdetails);
	}
      break;

    default:
      break;
    }
}


void
dump_ap_alert ()
{
  ap_stat_elem *current;
  long int period;
  float debit;
  current = ap_stat_list;

  while (current != NULL)
    {
      period =
	diffdate (current->data.t_last_packet, current->data.t_first_packet);
      if (period != 0)
	debit =
	  (float) current->data.num_offending_packet * 1000000 /
	  (float) period;
      else
	debit = 0;
      msg_alert (debit, current);
      current = current->next;
    }
}


ap_stat_elem *
find_ap_stat_elem (__u8 * ap_mac, __u8 * dst_mac, __u8 type)
{
  /* ap_mac, dst_mac & message_type are our "primary key" in the list */
  ap_stat_elem *current = ap_stat_list;

  while ((current != NULL) &&
	 ((memcmp (ap_mac, current->data.ap_mac, ETH_HW_ADDR_LEN) != 0) ||
	  (memcmp (dst_mac, current->data.dst_mac, ETH_HW_ADDR_LEN) != 0) ||
	  (type != current->data.type)))
    current = current->next;

  return current;
}

void
build_ap_stat ()
{
  ap_packet_elem *current;
  ap_stat_elem *actualize = NULL;
  ap_stat_elem *new;
  current = ap_packet_list;

  while (current != NULL)
    {
      actualize =
	find_ap_stat_elem (current->data.ap_mac, current->data.dst_mac,
			   current->data.type);
      if (actualize == NULL)
	{
	  /* add one */
	  new = (ap_stat_elem *) malloc (sizeof (ap_stat_elem));
	  memcpy (new->data.ap_mac, current->data.ap_mac, ETH_HW_ADDR_LEN);
	  memcpy (new->data.dst_mac, current->data.dst_mac, ETH_HW_ADDR_LEN);
	  new->data.type = current->data.type;
	  new->data.num_offending_packet = 1;
	  new->data.t_first_packet = current->data.packet_time;
	  new->data.t_last_packet = current->data.packet_time;
	  new->next = ap_stat_list;
	  ap_stat_list = new;
	}
      else
	{
	  actualize->data.num_offending_packet++;
	  actualize->data.t_first_packet =
	    min_date (current->data.packet_time,
		      actualize->data.t_first_packet);
	  actualize->data.t_last_packet =
	    max_date (current->data.packet_time,
		      actualize->data.t_last_packet);
	}
      current = current->next;
    }
}
