#include <asm/types.h>
#include <stdlib.h>
#include <unistd.h>
#include <syslog.h>
#include <time.h>
#include "usefull.h"
#include "packet80211b.h"
#include "card_session.h"

extern card_session_elem *card_session_list;
extern int syslog_alert;

card_session_elem *
find_card_session_elem (__u8 * card_mac)
{
  card_session_elem *current = card_session_list;
  while ((current != NULL)
	 && (memcmp (card_mac, current->data.card_mac, ETH_HW_ADDR_LEN) != 0))
    current = current->next;
  return current;
}

void
add_packet_card_session (packet80211b p)
{
  card_session_elem *new;
  new = find_card_session_elem (&p.packetdata[10]);

  if (new == NULL)
    {
      /* add one stat elem */
      new = (card_session_elem *) malloc (sizeof (card_session_elem));
      memcpy (new->data.card_mac, &p.packetdata[10], ETH_HW_ADDR_LEN);
      memcpy (new->data.ap_mac, &p.packetdata[16], ETH_HW_ADDR_LEN);
      new->data.t_first_packet = p.packet_time;
      new->data.t_last_packet = p.packet_time;
      new->data.t_flush = p.packet_time;
      new->data.num_auth = 0;
      new->data.num_assoc = 0;
      new->data.num_reassoc = 0;
      new->data.num_probe_request = 0;
      switch (p.packetdata[0])
	{
	case 0xb0:
	  new->data.num_auth++;
	  break;
	case 0x00:
	  new->data.num_assoc++;
	  break;
	case 0x20:
	  new->data.num_reassoc++;
	  break;
	case 0x40:
	  new->data.num_probe_request++;
	  break;
	default:
	  break;
	}
      new->data.num_data_sent = 0;
      new->data.num_data_received = 0;
      new->data.is_auth = 0;
      new->data.is_assoc = 0;
      new->data.WEP = 0;
      new->next = card_session_list;
      card_session_list = new;
    }
  else
    {
      /* actualize */
      if (diffdate (new->data.t_flush, new->data.t_last_packet) > 0)
	new->data.t_first_packet = p.packet_time;
      new->data.t_last_packet = p.packet_time;
      switch (p.packetdata[0])
	{
	case 0xb0:
	  new->data.num_auth++;
	  break;
	case 0x00:
	  new->data.num_assoc++;
	  break;
	case 0x20:
	  new->data.num_reassoc++;
	  break;
	case 0x40:
	  new->data.num_probe_request++;
	  break;
	default:
	  break;
	}

    }
}

void
dump_card_session ()
{
  card_session_elem *current;
  char *buf1;
  char *buf2;
  char *buf3;
  char *time1;
  char *time2;
  char *time3;
  long int period;
  float debit;


  current = card_session_list;
  buf1 = (char *) malloc (30 * sizeof (char));
  buf2 = (char *) malloc (30 * sizeof (char));
  buf3 = (char *) malloc (30 * sizeof (char));

  while (current != NULL)
    {
      /* REENTRANT safe */
      time1 = (char *) ctime_r (&current->data.t_first_packet.tv_sec, buf1);
      time2 = (char *) ctime_r (&current->data.t_last_packet.tv_sec, buf2);
      time3 = (char *) ctime_r (&current->data.t_flush.tv_sec, buf3);
      printf
	("\n-----------------\nCARD mac : %s\nAP mac : %s\nauth packets : %d\nassoc packets : %d\nreassoc packets : %d\nprobe request packets : %d\nfirst : %slast : %sflush : %s",
	 human_format_hw_addr (current->data.card_mac),
	 human_format_hw_addr (current->data.ap_mac), current->data.num_auth,
	 current->data.num_assoc, current->data.num_reassoc,
	 current->data.num_probe_request, time1, time2, time3);
      current = current->next;
      free (time1);
      free (time2);
    }
  free (buf1);
  free (buf2);
}


void
dump_card_alert ()
{
  card_session_elem *current;
  struct tm *tim1;
  struct tm *tim2;
  long int period;
  float debit;

  char *msgwho;
  char msgAP[50];
  char msgdetails[150];


  current = card_session_list;
  tim1 = (struct tm *) malloc (sizeof (struct tm));
  tim2 = (struct tm *) malloc (sizeof (struct tm));

  while (current != NULL)
    {
      /* REENTRANT safe */
      tim1 =
	(struct tm *) localtime_r (&current->data.t_first_packet.tv_sec,
				   tim1);
      tim2 =
	(struct tm *) localtime_r (&current->data.t_last_packet.tv_sec, tim2);
      msgwho = (char *) malloc (30 * sizeof (char));
      sprintf (msgwho, "CARD %s",
	       human_format_hw_addr (current->data.card_mac));
      if (isbroadcast (current->data.ap_mac))
	strcpy (msgAP, "broadcast");
      else
	sprintf (msgAP, "access point %s",
		 human_format_hw_addr (current->data.ap_mac));


      period =
	diffdate (current->data.t_last_packet, current->data.t_first_packet);
      if (period != 0)
	{

	  if ((float) current->data.num_auth * 1000000 / (float) period >
	      MAX_AUTH_REQUEST)
	    {
	      sprintf (msgdetails,
		       "(%d authentication requests sent to %s between %02d:%02d:%02d and %02d:%02d:%02d)",
		       current->data.num_auth, msgAP, tim1->tm_hour,
		       tim1->tm_min, tim1->tm_sec, tim2->tm_hour,
		       tim2->tm_min, tim2->tm_sec);
	      if (syslog_alert == 1)
		syslog (LOG_ALERT, "%s %s", msgwho, msgdetails);
	      else
		printf ("\n%s %s\n", msgwho, msgdetails);
	    }

	  if ((float) current->data.num_assoc * 1000000 / (float) period >
	      MAX_ASSOC_REQUEST)
	    {
	      sprintf (msgdetails,
		       "(%d association requests sent to %s between %02d:%02d:%02d and %02d:%02d:%02d)",
		       current->data.num_assoc, msgAP, tim1->tm_hour,
		       tim1->tm_min, tim1->tm_sec, tim2->tm_hour,
		       tim2->tm_min, tim2->tm_sec);
	      if (syslog_alert == 1)
		syslog (LOG_ALERT, "%s %s", msgwho, msgdetails);
	      else
		printf ("\n%s %s\n", msgwho, msgdetails);
	    }

	  if ((float) current->data.num_reassoc * 1000000 / (float) period >
	      MAX_REASSOC_REQUEST)
	    {
	      sprintf (msgdetails,
		       "(%d reassociation requests sent to %s between %02d:%02d:%02d and %02d:%02d:%02d)",
		       current->data.num_reassoc, msgAP, tim1->tm_hour,
		       tim1->tm_min, tim1->tm_sec, tim2->tm_hour,
		       tim2->tm_min, tim2->tm_sec);
	      if (syslog_alert == 1)
		syslog (LOG_ALERT, "%s %s", msgwho, msgdetails);
	      else
		printf ("\n%s %s\n", msgwho, msgdetails);
	    }

	  if ((float) current->data.num_probe_request * 1000000 /
	      (float) period > MAX_PROBE_REQUEST)
	    {
	      sprintf (msgdetails,
		       "(%d probe requests sent to %s between %02d:%02d:%02d and %02d:%02d:%02d)",
		       current->data.num_probe_request, msgAP, tim1->tm_hour,
		       tim1->tm_min, tim1->tm_sec, tim2->tm_hour,
		       tim2->tm_min, tim2->tm_sec);
	      if (syslog_alert == 1)
		syslog (LOG_ALERT, "%s %s", msgwho, msgdetails);
	      else
		printf ("\n%s %s\n", msgwho, msgdetails);
	    }
	}

      current = current->next;
    }
  free (tim1);
  free (tim2);
}

void
windowclean_card_session ()
{
  card_session_elem *current;
  struct timeval t_flush;
  gettimeofday (&t_flush, NULL);
  current = card_session_list;
  while (current != NULL)
    {
      current->data.t_flush = t_flush;
      current->data.num_auth = current->data.num_assoc = 0;
      current->data.num_reassoc = current->data.num_probe_request = 0;
      current->data.is_auth = current->data.is_assoc = current->data.WEP = 0;
      current = current->next;
    }

}
void
remove_card_session ()
{
  struct timeval t_now;
  card_session_elem *current;
  card_session_elem *precedent;
  int i = 0;
  gettimeofday (&t_now, NULL);
  while ((card_session_list != NULL) &&
	 (diffdate (t_now, card_session_list->data.t_last_packet) >
	  1000000 * WINDOW_CARD_RM))
    {
      current = card_session_list;
      card_session_list = current->next;
      free (current);
    }

  current = card_session_list;
  while (current != NULL)
    {
      precedent = current;
      current = current->next;
      if ((current != NULL)
	  && (diffdate (t_now, card_session_list->data.t_last_packet) >
	      1000000 * WINDOW_CARD_RM))
	{
	  precedent->next = current->next;
	  free (current);
	  current = precedent;
	}
    }
  current = card_session_list;
  /*
     while (current != NULL) {
     i++;
     current=current->next;
     }
     printf("\ncard_session_list : %d elements\n",i);
   */

}
