#include <asm/types.h>
#include <stdlib.h>

#define ETH_HW_ADDR_LEN 6

struct timeval
min_date (struct timeval t1, struct timeval t2)
{
  if (t1.tv_sec < t2.tv_sec)
    return t1;
  if ((t1.tv_sec == t2.tv_sec) && (t1.tv_usec < t2.tv_usec))
    return t1;
  return t2;
}

struct timeval
max_date (struct timeval t1, struct timeval t2)
{
  if (t1.tv_sec > t2.tv_sec)
    return t1;
  if ((t1.tv_sec == t2.tv_sec) && (t1.tv_usec > t2.tv_usec))
    return t1;
  return t2;
}

long int
diffdate (struct timeval t1, struct timeval t2)
{
  long int ret = (t1.tv_sec - t2.tv_sec) * 1000000 + t1.tv_usec - t2.tv_usec;
  return ret;
}

int
isbroadcast (__u8 * mac)
{
  __u8 broadcast[] = { 0xff, 0xff, 0xff, 0xff, 0xff, 0xff };
  if (memcmp (mac, broadcast, ETH_HW_ADDR_LEN) == 0)
    return 1;
  return 0;
}


void
hexdump (unsigned char *data, unsigned int amount)
/* canibalized from burneye/debug/memdump.c (scut) */
{
  unsigned int dp, p;		/* data pointer */
  const char trans[] =
    "................................ !\"#$%&'()*+,-./0123456789"
    ":;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`abcdefghijklm"
    "nopqrstuvwxyz{|}~...................................."
    "....................................................."
    "........................................";

  for (dp = 1; dp <= amount; dp++)
    {
      printf ("%02x ", data[dp - 1]);
      if ((dp % 8) == 0)
	printf (" ");
      if ((dp % 16) == 0)
	{
	  printf ("| ");
	  p = dp;
	  for (dp -= 16; dp < p; dp++)
	    printf ("%c", trans[data[dp]]);
	  printf ("\n");
	}
    }
  if ((amount % 16) != 0)
    {
      p = dp = 16 - (amount % 16);
      for (dp = p; dp > 0; dp--)
	{
	  printf ("  ");
	  if (((dp % 8) == 0) && (p != 8))
	    printf (" ");
	}
      printf (" | ");
      for (dp = (amount - (16 - p)); dp < amount; dp++)
	printf ("%c", trans[data[dp]]);
      printf ("%c", trans[data[dp]]);
    }
  printf ("\n");

  return;
}

char *
human_format_hw_addr (__u8 * str)
{
  char *buf;
  buf = (char *) malloc (sizeof (char) * 18);
  sprintf (buf, "%02x:%02x:%02x:%02x:%02x:%02x", str[0], str[1], str[2],
	   str[3], str[4], str[5]);
  return buf;
}

int
hexval (char ch)
{
  if (ch >= '0' && ch <= '9')
    return ch - '0';
  if (ch >= 'a' && ch <= 'f')
    return ch - 'a' + 10;
  if (ch >= 'A' && ch <= 'F')
    return ch - 'A' + 10;
  return -1;
}

u_char *
machine_format_key (char *str)
{
  u_char *buf;
  int i, h1, h2;
  char c, val;
  int len;

  if ((strlen (str) != 10) && (strlen (str) != 26))	/* 40 && 128 bit key */
    {
      perror ("Invalid key format");
      return 0;
    }

  len = strlen (str) / 2;
  buf = (u_char *) malloc (sizeof (u_char) * len);
  for (i = 0; i < len; i++)
    {
      h1 = hexval (*str++);
      h2 = hexval (*str++);

      if (h1 == -1 || h2 == -1)
	{
	  perror ("Invalid key format");
	  return 0;
	}

      buf[i] = (h1 << 4) | h2;
    }
  return buf;
}

void
get_hw_addr (char *buf, char *str)
{
/* stolen from arp_send.c (Yuri Volobuev) */
  int i;
  char c, val;
  for (i = 0; i < ETH_HW_ADDR_LEN; i++)
    {
      if (!(c = tolower (*str++)))
	perror ("Invalid hardware address");
      if (isdigit (c))
	val = c - '0';
      else if (c >= 'a' && c <= 'f')
	val = c - 'a' + 10;
      else
	perror ("Invalid hardware address");
      *buf = val << 4;
      if (!(c = tolower (*str++)))
	perror ("Invalid hardware address");
      if (isdigit (c))
	val = c - '0';
      else if (c >= 'a' && c <= 'f')
	val = c - 'a' + 10;
      else
	perror ("Invalid hardware address");
      *buf++ |= val;
      if (*str == ':')
	str++;
    }
}
