/************************************************************************
* Copyright (c) Mi Keli produccion, no 6				*
*									*
* This program is free software; you can redistribute it and/or modify	*
* it under the terms of the GNU General Public License version 2 as	*
* published by the Free Software Foundation. See COPYING for more	*
* details. 	 	 	 	 	 	 		*
* ===================================== 	 	 		*
*									* 
* wIDS : wireless IDS/honeypot 						*
*									*
* Usage : ./wIDS [-s] -i device [-h honeypot_MAC -l logfile ]	\	*
*		 [-o IDS_device]					*
*									*
* Ex	: ./wIDS -s -i eth1 -o eth0					*
*	  ./wIDS -i wlan0 -h 00:02:2d:20:34:1f -l clear.tcpdump		*
*									* 
* Note  : "-s" option should be used.					*
*									*
* What it can detect :							*
* 	- insistant probe requests (active sniffing)			*
*	- rude AP paquet sequences (802.11b Management frames jamming) 	*
*	- data trafic from/to unknowns MACs (...) 			*
*	- data trafic from/to unknowns APs (rogue APs)			*
*	- denied authentication						* 
*	- not encrypted trafic 						*
*									*
* Honeypot option :							*
*	- dump specific trafic (pcap like) on the honeypot access point *
*	  (should be without WEP, on a separate honeypot LAN)		*
* "-o" option :								*
*	- decrypt the trafic on our(s) AP(s) for level3 analysis	*
*									*
*************************************************************************/


#include <sys/types.h>
#include <asm/types.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <net/if.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <features.h>		/* for the glibc version number */
#include <netinet/in.h>
#if __GLIBC__ >= 2 && __GLIBC_MINOR >= 1
#include <netpacket/packet.h>
#include <net/ethernet.h>	/* the L2 protocols */
#else
#include <asm/types.h>
#include <linux/if_packet.h>
#include <linux/if_ether.h>	/* The L2 protocols */
#endif
#include <linux/types.h>
#include <syslog.h>
#include <limits.h>
#include <ctype.h>
#include <pthread.h>
#include "wep.h"
#include "usefull.h"
#include "packet80211b.h"
#include "ap_packet.h"
#include "card_session.h"
#include "ap_stat.h"
#include "beacon_interval.h"
#include "sequence_number.h"
#include "wepdata.h"
#include "wIDS.h"

/**************/
/* some globs */
/**************/

int syslog_alert = 0;
int flag_logfile = 0;
int flag_interface_IDS = 0;
int ids_sockfd;

char cap_ifname[IFNAMSIZ];
char cap_ofname[IFNAMSIZ];
__u8 inject_hw_address[ETH_HW_ADDR_LEN];

int num_known_card;
__u8 known_card[MAX_CARD_DEVICES][ETH_HW_ADDR_LEN];

int num_known_ap;
__u8 known_ap[MAX_AP_DEVICES][ETH_HW_ADDR_LEN];
char known_SSID[MAX_AP_DEVICES][33];

u_char *key[4];

__u8 honeypot[ETH_HW_ADDR_LEN];
int flag_honeypot = 0;

/*************************/
/* simple 802.11b packet */
/*************************/
packet80211b_elem *packet_list = NULL;
pthread_mutex_t mut_packet_list;



/*******************************************************
* Access points packets, cleaned every WINDOW_TIME_AP  *
********************************************************/
ap_packet_elem *ap_packet_list = NULL;
pthread_mutex_t mut_ap_list;


/********
* stats *
********/
ap_stat_elem *ap_stat_list = NULL;
pthread_mutex_t mut_ap_stat;

card_session_elem *card_session_list = NULL;
pthread_mutex_t mut_card_session;

/***********************************************
* sequence number & beacon intervals reminders *
***********************************************/
int seq_num_alert;
sequence_number_elem *sequence_number_list = NULL;
pthread_mutex_t mut_sequence_number;

int beacon_alert;
beacon_interval_elem *beacon_interval_list = NULL;
pthread_mutex_t mut_beacon_interval;

/********
* event *
********/
int denied_auth_alert;

/************
* dump file *
************/
char logfile[256];
__u8 logbuffer[LOGBUFFER];
int logpos = 0;
pthread_mutex_t mut_log;

/*************
* some funcs *
**************/
void
int_handler (int sig)
{
  char c;
  signal (sig, SIG_IGN);
  printf ("\nDo you really want to quit? [y/n]\n");
  c = getchar ();
  if (c == 'y' || c == 'Y')
    exit (0);
  else
    signal (SIGINT, int_handler);
}


void
writelog (packet80211b p)
{
  FILE *fd;
  if ((logpos + p.length) > LOGBUFFER)
    {
      pthread_mutex_lock (&mut_log);
      /* we write the current buffer in log file */
      if ((fd = fopen (logfile, "a")) == NULL)
	{
	  perror ("fopen logfile");
	  return;
	}
      fwrite (&p.packet_time, sizeof (struct timeval), 1, fd);
      fwrite (&p.length, sizeof (__u32), 1, fd);
      fwrite (&p.length, sizeof (__u32), 1, fd);
      fwrite (&p.packetdata, sizeof (__u8), p.length, fd);
      fclose (fd);
      pthread_mutex_unlock (&mut_log);
    }
  else
    {
      pthread_mutex_lock (&mut_log);
      memcpy (&logbuffer[logpos], p.packetdata, p.length);
      logpos += p.length;
      pthread_mutex_unlock (&mut_log);
    }
  return;
}


static void
usage (void)
{
  fprintf (stderr,
	   "\nusage : ./wIDS [-s] -i device [-l logfile -h honeypot] [-o device]\n\n"
	   "options :\n"
	   "          -s          : use syslog (LOG_ALERT)\n"
	   "          -i device   : listen on the interface specified by device (eth0, wlan0...)\n"
	   "                        (should be in promiscuous mode)\n"
	   "          -l logfile  : file where honeypot packets 'll be dumped\n"
	   "          -h honeypot : alert about trafic on the specified honeypot AP' MAC\n"
	   "          -o device   : device where decrypted trafic is sent for IDS analisis\n\n"
	   "note    : \"-s\" option should be used.\n\n"
	   "exemple : ./wIDS -s -i eth1 -o eth0\n"
	   "          ./wIDS -s -i wlan0 -l ./wIDS.tcpdump -h 00:02:2d:4b:7e:0a\n\n");
  exit (1);
}


static int
is_known_card (__u8 * mac)
{
  int i;
  for (i = 0; i < num_known_card; i++)
    if (memcmp (mac, known_card[i], ETH_HW_ADDR_LEN) == 0)
      return 1;
  return 0;
}

static int
is_known_ap (__u8 * mac)
{
  int i;
  for (i = 0; i < num_known_ap; i++)
    if (memcmp (mac, known_ap[i], ETH_HW_ADDR_LEN) == 0)
      return 1;
  return 0;
}

static int
is_honeypot (__u8 * mac)
{
  if (memcmp (mac, honeypot, ETH_HW_ADDR_LEN) == 0)
    return 1;
  return 0;
}

void
actualize_is_known_card ()
{
  card_session_elem *current;

  current = card_session_list;
  while (current != NULL)
    {
      if (is_known_card (current->data.card_mac))
	current->data.is_known = 1;
      else
	current->data.is_known = 0;
      current = current->next;
    }
}

static void
get_injector_hw_device (char *ifname)
{
  unsigned char *ch;
  int i;
  int fd;
  struct ifreq ifr;

  fd = socket (PF_INET, SOCK_PACKET, htons (ETH_P_ALL));	/* open socket */
  strcpy (ifr.ifr_name, ifname);
  ioctl (fd, SIOCGIFHWADDR, &ifr);	/* retrieve MAC address */
  memcpy (inject_hw_address, ifr.ifr_hwaddr.sa_data, ETH_HW_ADDR_LEN);
  close (fd);
}


static int
get_raw_socket (char *ifname)
/* ripped from airjack/tools (Abaddon) */
{
  struct sockaddr_ll addr;
  struct ifreq req;
  int sockfd;


  /* open the link layer socket */
  if ((sockfd = socket (PF_PACKET, SOCK_RAW, htons (ETH_P_ALL))) < 0)
    {
      return (-1);
    }

  /* get the interface index */
  memset (&req, 0, sizeof (struct ifreq));
  strcpy (req.ifr_name, ifname);

  if (ioctl (sockfd, SIOCGIFINDEX, &req) < 0)
    {
      return (-2);
    }

  /* bind the socket to the interface */
  memset (&addr, 0, sizeof (struct sockaddr_ll));
  addr.sll_ifindex = req.ifr_ifindex;
  addr.sll_protocol = htons (ETH_P_ALL);
  addr.sll_family = AF_PACKET;
  if (bind (sockfd, (struct sockaddr *) &addr, sizeof (struct sockaddr_ll)) <
      0)
    {
      return (-3);
    }

  return (sockfd);
}

void *
wIDSheel ()
{
  char wheel[4] = { '|', '/', '-', '\\' };
  int i = 0;
  setbuf (stdout, NULL);
  while (1)
    {
      usleep (75000);
      printf ("\rwIDS is sniffing for hostile trafic... %c",
	      wheel[(++i) % 4]);
    }
}

/*****************
* decrypt thread *
*****************/
void *
decrypt (void *arg)
{
  __u8 *ethernet_packet;
  wepdata *packet;
  int datalen;

  packet = (wepdata *) arg;
  datalen = packet->length;

  wep_decrypt (packet->packetdata, datalen, packet->key, 0, STRONG);

  ethernet_packet = (__u8 *) malloc (sizeof (__u8) * (datalen + 2));
  memcpy (ethernet_packet, packet->dst, ETH_HW_ADDR_LEN);

  /* the original MAC sender is changed to our device one    */
  /* memcpy (ethernet_packet, packet->src, ETH_HW_ADDR_LEN); */
  memcpy (&ethernet_packet[ETH_HW_ADDR_LEN], inject_hw_address,
	  ETH_HW_ADDR_LEN);

  memcpy (&ethernet_packet[2 * ETH_HW_ADDR_LEN], &packet->packetdata[10],
	  datalen - 10);
  if (write (ids_sockfd, ethernet_packet, datalen + 2) != datalen + 2)
    {
      if ((errno == EWOULDBLOCK) || (errno == EINTR))
	{
	  perror ("write() wouldblock o EINTR on output ids device.");
	}
      else
	{
	  perror ("Fatal error while sending to output ids device.");
	  exit (1);
	}
    }
  free (ethernet_packet);
  //pthread_exit(NULL);
}

/*******************
* cleaning threads *
*******************/
void *
clean_card_session ()
{
  while (1)
    {
      sleep (WINDOW_TIME_CARD);
      pthread_mutex_lock (&mut_card_session);
      windowclean_card_session ();
      pthread_mutex_unlock (&mut_card_session);
    }
  /* cleanup, but never gets here  */
  pthread_exit (NULL);
}

/************
* data loop *
************/

void
data_loop (packet80211b p)
{
/* must be light for performances */

  int i;
  int from_known_card = 0;
  int to_known_card = 0;
  int from_known_ap = 0;
  int to_known_ap = 0;
  int by_known_ap = 0;
  int is_honeypot_trafic = 0;
  int is_wep = 1;
  __u8 src_mac[ETH_HW_ADDR_LEN];
  __u8 ap_mac[ETH_HW_ADDR_LEN];
  __u8 dst_mac[ETH_HW_ADDR_LEN];
  __u8 card1[ETH_HW_ADDR_LEN];
  __u8 card2[ETH_HW_ADDR_LEN];
  __u8 ap[ETH_HW_ADDR_LEN];
  __u16 addr_type;
  __u16 fc;
  wepdata *new;
  pthread_t decrypt_thread;
  //__u8* data;
  //__u8* ethernet_packet;
  /* protocols dumped are ARP, rARP, IP, IPX, IP6 and PPP       */
  /* source :  http://www.iana.org/assignments/ethernet-numbers */
  __u16 known_protocol[] = { 0x0806, 0x0800, 0x8037, 0x86dd, 0x8035, 0x880b };
  char msg[100];



  if ((p.packetdata[1] ^ 0x40) > 0x0F)
    is_wep = 0;

  /* avoiding manufacturers packets... */
  if (!(is_wep))
    for (i = 0; i < 5; i++)
      {
	if ((p.packetdata[30] * 16 * 16 + p.packetdata[31]) ==
	    known_protocol[i])
	  break;
	if (i == 4)
	  return;		/* not analized packet */
      }



  /* data_addr type dissector from ethereal (Gerald Combs & co.) */
  memcpy (&fc, p.packetdata, 2);
  addr_type = COOK_ADDR_SELECTOR (fc);
  /* In order to show src/dst address we must always do the following */
  switch (addr_type)
    {
    case DATA_ADDR_T1:
      memcpy (dst_mac, &p.packetdata[4], ETH_HW_ADDR_LEN);
      memcpy (src_mac, &p.packetdata[10], ETH_HW_ADDR_LEN);
      memcpy (ap_mac, &p.packetdata[16], ETH_HW_ADDR_LEN);
      break;

    case DATA_ADDR_T2:
      memcpy (dst_mac, &p.packetdata[4], ETH_HW_ADDR_LEN);
      memcpy (src_mac, &p.packetdata[16], ETH_HW_ADDR_LEN);
      memcpy (ap_mac, &p.packetdata[10], ETH_HW_ADDR_LEN);
      break;

    case DATA_ADDR_T3:
      memcpy (dst_mac, &p.packetdata[16], ETH_HW_ADDR_LEN);
      memcpy (src_mac, &p.packetdata[10], ETH_HW_ADDR_LEN);
      memcpy (ap_mac, &p.packetdata[4], ETH_HW_ADDR_LEN);
      break;

    case DATA_ADDR_T4:
      memcpy (dst_mac, &p.packetdata[16], ETH_HW_ADDR_LEN);
      memcpy (src_mac, &p.packetdata[24], ETH_HW_ADDR_LEN);
      memcpy (ap_mac, &p.packetdata[4], ETH_HW_ADDR_LEN);
      break;
    default:
      printf ("\n failed to get hddr \n");
      return;
      break;
    }

  if (is_known_card (src_mac))
    {
      from_known_card = 1;
      memcpy (card1, src_mac, ETH_HW_ADDR_LEN);
    }
  if (is_known_card (dst_mac))
    {
      to_known_card = 1;
      memcpy (card2, src_mac, ETH_HW_ADDR_LEN);
    }
  if (is_known_ap (dst_mac))
    {
      to_known_ap = 1;
      memcpy (ap, dst_mac, ETH_HW_ADDR_LEN);
    }
  if (is_known_ap (src_mac))
    {
      from_known_ap = 1;
      memcpy (ap, src_mac, ETH_HW_ADDR_LEN);
    }
  if (is_known_ap (ap_mac))
    by_known_ap = 1;
  if (is_honeypot (src_mac) || is_honeypot (dst_mac) || is_honeypot (ap_mac))
    is_honeypot_trafic = 1;


  if (is_honeypot_trafic)
    {
      strcpy (msg, "[ALERT] Honeypot data trafic detected!");
      if (syslog_alert)
	syslog (LOG_ALERT, "%s", msg);
      else
	printf ("\n%s\n", msg);
      if (flag_logfile)
	writelog (p);
    }


  if ((from_known_ap || to_known_ap || by_known_ap)
      && (from_known_card || to_known_card))
    {
      if (is_wep)
	{
	  strcpy (msg,
		  "[INFO] WEP data trafic from/to known client to/from/by known access point");
	  if (syslog_alert)
	    syslog (LOG_ALERT, "%s", msg);
	  else
	    printf ("\n%s\n", msg);
	}
      else
	{
	  strcpy (msg,
		  "[ALERT] Clear data trafic from/to known client to/from/by known access point!");
	  if (syslog_alert)
	    syslog (LOG_ALERT, "%s", msg);
	  else
	    printf ("\n%s\n", msg);
	  if (flag_logfile)
	    writelog (p);
	}
    }

  else if ((from_known_ap || to_known_ap || by_known_ap)
	   && !(from_known_card || to_known_card))
    {
      if (is_wep)
	{
	  strcpy (msg,
		  "[ALERT] WEP data trafic from/to unknown client to/from known access point!");
	  if (syslog_alert)
	    syslog (LOG_ALERT, "%s", msg);
	  else
	    printf ("\n%s\n", msg);
	}
      else
	{
	  strcpy (msg,
		  "[ALERT] Clear data trafic from/to unknown client to/from known access point!");
	  if (syslog_alert)
	    syslog (LOG_ALERT, "%s", msg);
	  else
	    printf ("\n%s\n", msg);
	  if (flag_logfile)
	    writelog (p);
	}
    }
  else if (!(from_known_card || to_known_card)
	   && !(from_known_ap || to_known_ap || by_known_ap))
    {
      if (is_wep)
	{
	  strcpy (msg,
		  "[WARNING] WEP data trafic detected between unknowns card and access point!");
	  if (syslog_alert)
	    syslog (LOG_ALERT, "%s", msg);
	  else
	    printf ("\n%s\n", msg);
	}
      else
	{
	  strcpy (msg,
		  "[WARNING] Clear data trafic detected between unknowns card and access point!");
	  if (syslog_alert)
	    syslog (LOG_ALERT, "%s", msg);
	  else
	    printf ("\n%s\n", msg);
	  /*if (flag_logfile) */
	  /*  writelog (p);   */
	}
    }
  else if ((from_known_card || to_known_card)
	   && !(from_known_ap || to_known_ap || by_known_ap))
    {
      if (is_wep)
	{
	  strcpy (msg,
		  "[ALERT] WEP data trafic from/to known client to/from unknown access point!");
	  if (syslog_alert)
	    syslog (LOG_ALERT, "%s", msg);
	  else
	    printf ("\n%s\n", msg);
	}
      else
	{
	  strcpy (msg,
		  "[ALERT] Clear data trafic from/to known client to/from unknown access point!");
	  if (syslog_alert)
	    syslog (LOG_ALERT, "%s", msg);
	  else
	    printf ("\n%s\n", msg);
	}
      if (flag_logfile)
	writelog (p);
    }

  /* send decrypted wep packet to output device if provided */
  //if (is_wep && flag_interface_IDS && to_known_card && (from_known_ap || by_known_ap))
  if (is_wep && flag_interface_IDS)
    {
      new = (wepdata *) malloc (sizeof (wepdata));
      new->packetdata = (__u8 *) malloc (sizeof (__u8) * (p.length - 24));
      new->key = (__u8 *) malloc (sizeof (__u8) * LENKEY (STRONG));
      memcpy (new->packetdata, &p.packetdata[24], p.length - 24);
      new->length = p.length - 24;
      memcpy (new->key, key[p.packetdata[27]], LENKEY (STRONG));

      memcpy (new->src, src_mac, ETH_HW_ADDR_LEN);
      memcpy (new->dst, dst_mac, ETH_HW_ADDR_LEN);
      if (decrypt (new) < 0)
	perror ("error while decrypted trafic injection");
      //if (pthread_create(&decrypt_thread, NULL, decrypt, (void*)new) < 0)
      //      perror("pthread_create decrypt_thread");

      /* cleanup */
      free (new->packetdata);
      free (new->key);
      free (new);
    }
}


/*************************
* info gathering threads *
*************************/

void
actualize_card_session (packet80211b p)
{
  pthread_mutex_lock (&mut_card_session);
  add_packet_card_session (p);
  pthread_mutex_unlock (&mut_card_session);
}

void
add_ap_packet_list (packet80211b p)
{
  pthread_mutex_lock (&mut_ap_list);
  add_ap_packet (p);
  pthread_mutex_unlock (&mut_ap_list);
}

/*********
* alerts *
*********/

void
dump_event_alerts ()
{
  if (seq_num_alert != 0)
    if (syslog_alert)
      syslog (LOG_ALERT, "Bad sequence numbers : %d", seq_num_alert);
    else
      printf ("\nBad sequence numbers : %d\n", seq_num_alert);

  if (beacon_alert != 0)
    if (syslog_alert)
      syslog (LOG_ALERT, "Invalid beacon intervals : %d", beacon_alert);
    else
      printf ("\nInvalid beacon intervals : %d\n", beacon_alert);
}

void
actualize_beacon_alert (packet80211b p)
{
  /* we suppose p is beacon type */
  int new_beacon_interval;
  int old_beacon_interval;
  /* beacon change */
  new_beacon_interval = (p.packetdata[32] + p.packetdata[33] * 16) * 1024;
  pthread_mutex_lock (&mut_beacon_interval);
  old_beacon_interval = beacon_interval_val (p);
  pthread_mutex_unlock (&mut_beacon_interval);
  if (old_beacon_interval == -1)
    {
      pthread_mutex_lock (&mut_beacon_interval);
      add_beacon_interval (p);
      pthread_mutex_unlock (&mut_beacon_interval);
    }
  else if (old_beacon_interval != new_beacon_interval)
    {
      pthread_mutex_lock (&mut_beacon_interval);
      beacon_alert++;
      pthread_mutex_unlock (&mut_beacon_interval);
    }
}


void
actualize_sequence_alert (packet80211b p)
{
  /* we suppose p is beacon o ap_mgt */
  /* curious packet sequence number  */

  pthread_mutex_lock (&mut_sequence_number);
  if (curious_seq_num (p))
    seq_num_alert++;
  pthread_mutex_unlock (&mut_sequence_number);

}


void
alert_event_denied_auth (packet80211b p)
{
  /* we supose p is ap_mgt */
  if (is_negative_auth (p))
    if (syslog_alert)
      syslog (LOG_ALERT, "Denied Auth from access point %s to card %s",
	      human_format_hw_addr (&p.packetdata[16]),
	      human_format_hw_addr (&p.packetdata[4]));
    else
      printf ("\nDenied Auth from access point %s to card %s\n",
	      human_format_hw_addr (&p.packetdata[16]),
	      human_format_hw_addr (&p.packetdata[4]));

}


void
alert_wellenreiter ()
{

  /* wellenreiter spoof the card MAC to emit probe request    */
  /* so we look how much sessions are opened and global probe */
  /* request frames sent.                                     */
  /* Ho.. I didn't tested it :>                               */

  float debit = 0;
  int num_unknown_cards = 0;
  int num_probe_requests = 0;
  card_session_elem *current;

  current = card_session_list;
  while (current != NULL)
    {
      if (!(current->data.is_known))
	{
	  num_unknown_cards++;
	  num_probe_requests += current->data.num_probe_request;
	  current = current->next;
	}
    }

  if (num_unknown_cards > 4)
    debit = (float) num_probe_requests / (float) WINDOW_TIME_AP;

  if (debit > MAX_PROBE_REQUEST)
    if (syslog_alert)
      syslog (LOG_ALERT,
	      "[ALERT] Wellenreiter-like active sniffing detected");
}


/*************
* dispatcher *
*************/
void *
dispatch ()
{
  packet80211b p;
  packet80211b_elem *current;
  int new_beacon_interval;
  int old_beacon_interval;
  while (1)
    {
      pthread_mutex_lock (&mut_packet_list);
      current = packet_list;
      if (current == NULL)
	{
	  pthread_mutex_unlock (&mut_packet_list);
	  usleep (10000);
	}
      else
	{
	  /* p is the first packet of the list */
	  memcpy (p.packetdata, current->data.packetdata,
		  current->data.length);
	  p.packet_time = current->data.packet_time;
	  p.length = current->data.length;
	  packet_list = current->next;

	  free (current);
	  pthread_mutex_unlock (&mut_packet_list);

	  /* dispatch p depending on its type */
	  if (is_data (p))
	    data_loop (p);
	  else if (is_card_mgt (p))
	    actualize_card_session (p);
	  else if (is_beacon (p))
	    {
	      actualize_beacon_alert (p);
	      actualize_sequence_alert (p);
	      add_ap_packet_list (p);
	    }
	  else if (is_ap_mgt (p))
	    {
	      alert_event_denied_auth (p);
	      actualize_sequence_alert (p);
	      add_ap_packet_list (p);
	    }
	}
    }				/* end while */
/* cleanup, but never gets here */
  pthread_exit (NULL);
}

/*****************
* capture thread *
*****************/
void *
capture ()
{
  int i, cap_sockfd;
  ssize_t nbytes;
  __u8 frame_buf[4096];		/* capture_buffer */
  packet80211b p;


  /* read from the device socket */
  if ((cap_sockfd = get_raw_socket (cap_ifname)) < 0)
    {
      perror ("get_raw_socket");
      exit (1);
    }

  /* mutex initialization */
  pthread_mutex_init (&mut_packet_list, NULL);

  while (1)
    {
      memset (frame_buf, 0, sizeof (frame_buf));
      if ((nbytes = read (cap_sockfd, frame_buf, sizeof (frame_buf))) < 0)
	{
	  if ((errno == EWOULDBLOCK) || (errno == EINTR))
	    {
	      continue;		/* blah Urself */
	    }
	  perror ("Fatal error while sniffing.");
	  exit (1);
	}
      gettimeofday (&p.packet_time, NULL);
      memcpy (p.packetdata, frame_buf, nbytes);
      p.length = nbytes;

      /* adding packet in list */
      pthread_mutex_lock (&mut_packet_list);
      add_packet80211b (p, packet_list);
      pthread_mutex_unlock (&mut_packet_list);
    }				/* while */
  /* cleanup, but never gets here */
  pthread_exit (NULL);
}



/********************************************************
*			main				*
********************************************************/
int
main (int argc, char **argv)
{
  FILE *ap_file;
  FILE *card_file;
  FILE *key_file;
  FILE *fd_hdr;
  int cap_sockfd;
  int i;
  char line[80];
  char *ptr;
  struct timeval t_clean;
  pthread_t capture_thread;
  pthread_t dispatch_thread;
  pthread_t clean_thread_ap;
  pthread_t clean_thread_session;
  pthread_t wheel_thread;
  char *valid_options;
  int flag_interface = 0;
  int num_wep_keys = 0;
  int ch;			/* storage var for getopt info */
  pcap_header hdr;

  signal (SIGINT, int_handler);	/* CTRL-C handler */
  valid_options = "sh:i:l:o:";

  while ((ch = getopt (argc, argv, valid_options)) != -1)
    {
      switch (ch)
	{
	case 's':		/* syslog */
	  syslog_alert = 1;
	  break;
	case 'i':		/* wireless interface */
	  strncpy (cap_ifname, optarg, IFNAMSIZ);
	  flag_interface = 1;
	  break;
	case 'h':		/* honeypot MAC */
	  flag_honeypot = 1;
	  get_hw_addr (honeypot, optarg);
	  break;
	case 'l':		/* dump */
	  strncpy (logfile, optarg, 255);
	  creat (logfile, S_IRWXU);
	  /* we write an adecuate pcap header in log file */
	  hdr.magic = 0xa1b2c3d4;
	  hdr.major = 2;
	  hdr.minor = 4;
	  hdr.thiszone = 0;
	  hdr.timestamp_accuracy = 0;
	  hdr.snaplen = 0x0000FFFF;
	  hdr.linktype = 105;	/* 802.11 */
	  if ((fd_hdr = fopen (logfile, "a")) == NULL)
	    {
	      perror ("fopen logfile");
	      return;
	    }
	  if (fwrite (&hdr, sizeof (pcap_header), 1, fd_hdr) <= 0)
	    {
	      perror ("fwrite logfile");
	      return;
	    }
	  fclose (fd_hdr);
	  flag_logfile = 1;
	  break;
	case 'o':		/* output interface */
	  strncpy (cap_ofname, optarg, IFNAMSIZ);

	  /* get MAC to avoid ARP stress (this work for eth devices) */
	  get_injector_hw_device (cap_ofname);

	  /* create the device socket */
	  if ((ids_sockfd = get_raw_socket (cap_ofname)) < 0)
	    {
	      perror ("output dev get_raw_socket");
	      exit (1);
	    }

	  flag_interface_IDS = 1;
	  key_file = fopen ("./Keys.lst", "r");
	  if (key_file)
	    {
	      i = 0;
	      while (fgets (line, 80, key_file))
		{

		  num_wep_keys++;
		  if (num_wep_keys > 4)
		    {
		      perror ("Too much wep keys in Keys.lst");
		      return;
		    }
		  key[i] = machine_format_key (strtok (line, " \n"));
		  i++;
		}

	      fclose (key_file);
	    }
	  else
	    {
	      perror ("Create a Keys.lst file for -o option");
	      return;
	    }
	  break;

	default:
	  printf ("Unknown option -%c\n", ch);
	  usage ();
	  break;
	}
    }

  if (!flag_interface)
    usage ();


  if ((cap_sockfd = get_raw_socket (cap_ifname)) < 0)
    {
      perror ("get_raw_socket");
      exit (1);
    }

  system ("clear");
  printf ("\n----------> Starting wIDS on %s <----------\n\n", cap_ifname);

  if (syslog_alert)
    syslog (LOG_ALERT, "Starting wIDS...");

  i = 0;
  ap_file = fopen ("./KnownAPs.lst", "r");
  if (ap_file)
    {
      while (fgets (line, 80, ap_file))
	{
	  get_hw_addr (known_ap[i], strtok (line, " \n"));
	  ptr = strtok (NULL, " \n");
	  strcpy (known_SSID[i], ptr);
	  known_SSID[i][33] = '\0';
	  i++;
	}
      fclose (ap_file);
    }
  num_known_ap = i;

  printf ("known ap(s):\n");
  for (i = 0; i < num_known_ap; i++)
    printf ("MAC : %s\t\tSSID: %s\n", human_format_hw_addr (known_ap[i]),
	    known_SSID[i]);


  i = 0;
  card_file = fopen ("./KnownCards.lst", "r");
  if (card_file)
    {
      while (fgets (line, 80, card_file))
	{
	  get_hw_addr (known_card[i], strtok (line, " \n"));
	  i++;
	}
      fclose (card_file);
    }
  num_known_card = i;
  printf ("\nknown card(s):\n");
  for (i = 0; i < num_known_card; i++)
    printf ("MAC : %s\n", human_format_hw_addr (known_card[i]));

  if (flag_honeypot)
    printf ("\nHoneypot_ap : %s\n", human_format_hw_addr (honeypot));


  /* initialize mutexs */
  pthread_mutex_init (&mut_packet_list, NULL);
  pthread_mutex_init (&mut_ap_list, NULL);
  pthread_mutex_init (&mut_ap_stat, NULL);
  pthread_mutex_init (&mut_card_session, NULL);
  pthread_mutex_init (&mut_sequence_number, NULL);
  pthread_mutex_init (&mut_beacon_interval, NULL);
  pthread_mutex_init (&mut_log, NULL);

  if (pthread_create (&wheel_thread, NULL, wIDSheel, NULL) < 0)
    {
      if (errno == EINTR)
	{			/* rulzz  */
	}
    }

  /* launching capture thread */
  if (pthread_create (&capture_thread, NULL, capture, NULL) < 0)
    perror ("pthread_create capture_thread");

  /* firing dispatchers */
  for (i = 0; i < NUM_DISPATCHER; i++)
    {
      if (pthread_create (&dispatch_thread, NULL, dispatch, NULL) < 0)
	perror ("pthread_create dispatch");
    }

  /* card sessions cleaning threads */
  if (pthread_create (&clean_thread_session, NULL, clean_card_session, NULL) <
      0)
    perror ("pthread_create clean_thread_session");


  printf ("\n");
  setbuf (stdout, NULL);

  /* main loop */
  while (1)
    {
      gettimeofday (&t_clean, NULL);
      seq_num_alert = 0;
      beacon_alert = 0;
      denied_auth_alert = 0;
      sleep (WINDOW_TIME_AP);


      /* clean ap_list */
      pthread_mutex_lock (&mut_ap_list);
      flush_ap_packet (t_clean);
      /* reinitialize sequence numbers */
      pthread_mutex_lock (&mut_sequence_number);
      reset_sequence_number ();
      pthread_mutex_unlock (&mut_sequence_number);
      /* now we can build ap_stat, */
      flush_ap_stat ();
      build_ap_stat ();
      /* dump event alerts... */
      dump_event_alerts ();
      /* dump ap stat alerts... */
      dump_ap_alert ();
      pthread_mutex_unlock (&mut_ap_list);


      pthread_mutex_lock (&mut_card_session);
      /* we remove old list elems */
      remove_card_session ();
      /* actualize known/unknown cards in the list */
      actualize_is_known_card ();
      /* detect wellenreiter-like sniffing */
      alert_wellenreiter ();
      /* and dump session alerts */
      dump_card_alert ();
      pthread_mutex_unlock (&mut_card_session);
    }

/* never gets here */
  pthread_exit (NULL);
  return 0;
}
